@@ -37,6 +37,11 @@
 #error "Unsupported postgresql version"
 #endif
 
+#if PG_VERSION_NUM > 90299
+#include "access/htup_details.h" /* 9.3 heap_form_tuple */
+#include "common/relpath.h" /* 9.3 relpathbackend */
+#endif
+
 #ifdef PG_MODULE_MAGIC
 PG_MODULE_MAGIC;
 #endif
@@ -135,6 +140,9 @@ static int	pgfincore_file(char *filename, pgfincoreStruct *pgfncr);
  * and support 8.3, 8.4 and 9.0
  */
 #if PG_MAJOR_VERSION == 803
+#if defined(HAVE_DECL_POSIX_FADVISE)
+#define USE_POSIX_FADVISE
+#endif
 char *text_to_cstring(const text *t);
 text *cstring_to_text(const char *s);
 text *cstring_to_text_with_len(const char *s, int len);
@@ -225,20 +233,22 @@ pgsysconf(PG_FUNCTION_ARGS)
 	PG_RETURN_DATUM( HeapTupleGetDatum(tuple) );
 }
 
-#ifdef USE_POSIX_FADVISE
+#if defined(USE_POSIX_FADVISE)
 /*
  * pgfadvise_file
  */
 static int
-pgfadvise_file(char *filename, int advice, pgfadviseStruct	*pgfdv)
+pgfadvise_file(char *filename, int advice, pgfadviseStruct *pgfdv)
 {
 	/*
-	 * We work directly with the file
-	 * we don't use the postgresql file handler
+	 * We use the AllocateFile(2) provided by PostgreSQL.  We're going to
+	 * close it ourselves even if PostgreSQL close it anyway at transaction
+	 * end.
 	 */
-	struct stat	st;
-	int			fd;
-	int			adviceFlag;
+	FILE	*fp;
+	int	fd;
+	struct stat st;
+	int	    adviceFlag;
 
 	/*
 	 * OS Page size and Free pages
@@ -246,16 +256,18 @@ pgfadvise_file(char *filename, int advice, pgfadviseStruct	*pgfdv)
 	pgfdv->pageSize	= sysconf(_SC_PAGESIZE);
 
 	/*
-	 * Open and fstat file
+	 * Fopen and fstat file
 	 * fd will be provided to posix_fadvise
 	 * if there is no file, just return 1, it is expected to leave the SRF
 	 */
-	fd = open(filename, O_RDONLY);
-	if (fd == -1)
-		return 1;
+	fp = AllocateFile(filename, "rb");
+	if (fp == NULL)
+                return 1;
+
+	fd = fileno(fp);
 	if (fstat(fd, &st) == -1)
 	{
-		close(fd);
+		FreeFile(fp);
 		elog(ERROR, "pgfadvise: Can not stat object file : %s", filename);
 		return 2;
 	}
@@ -314,7 +326,7 @@ pgfadvise_file(char *filename, int advice, pgfadviseStruct	*pgfdv)
 	posix_fadvise(fd, 0, 0, adviceFlag);
 
 	/* close the file */
-	close(fd);
+	FreeFile(fp);
 
 	/*
 	 * OS things : Pages free
@@ -478,7 +490,7 @@ pgfadvise(PG_FUNCTION_ARGS)
 	}
 }
 
-#ifdef USE_POSIX_FADVISE
+#if defined(USE_POSIX_FADVISE)
 /*
  * pgfadvise_file
  */
@@ -493,11 +505,13 @@ pgfadvise_loader_file(char *filename,
 	int		i, k;
 
 	/*
-	 * We work directly with the file
-	 * we don't use the postgresql file handler
+	 * We use the AllocateFile(2) provided by PostgreSQL.  We're going to
+	 * close it ourselves even if PostgreSQL close it anyway at transaction
+	 * end.
 	 */
-	struct stat	st;
-	int			fd;
+	FILE	*fp;
+	int	fd;
+	struct stat st;
 
 	/*
 	 * OS things : Page size
@@ -513,16 +527,18 @@ pgfadvise_loader_file(char *filename,
 	pgfloader->pagesUnloaded	= 0;
 
 	/*
-	 * Open and fstat file
+	 * Fopen and fstat file
 	 * fd will be provided to posix_fadvise
 	 * if there is no file, just return 1, it is expected to leave the SRF
 	 */
-	fd = open(filename, O_RDONLY);
-	if (fd == -1)
-		return 1;
+	fp = AllocateFile(filename, "rb");
+	if (fp == NULL)
+                return 1;
+
+	fd = fileno(fp);
 	if (fstat(fd, &st) == -1)
 	{
-		close(fd);
+		FreeFile(fp);
 		elog(ERROR, "pgfadvise_loader: Can not stat object file: %s", filename);
 		return 2;
 	}
@@ -591,7 +607,7 @@ pgfadvise_loader_file(char *filename,
 			x <<= 1;
 		}
 	}
-	close(fd);
+	FreeFile(fp);
 
 	/*
 	 * OS things : Pages free
@@ -626,7 +642,7 @@ pgfadvise_loader(PG_FUNCTION_ARGS)
 	int       segmentNumber = PG_GETARG_INT32(2);
 	bool      willneed      = PG_GETARG_BOOL(3);
 	bool      dontneed      = PG_GETARG_BOOL(4);
-	VarBit    *databit		= PG_GETARG_VARBIT_P(5);
+	VarBit    *databit;
 
 	/* our structure use to return values */
 	pgfloaderStruct	*pgfloader;
@@ -646,6 +662,11 @@ pgfadvise_loader(PG_FUNCTION_ARGS)
 	Datum		values[PGFADVISE_LOADER_COLS];
 	bool		nulls[PGFADVISE_LOADER_COLS];
 
+	if (PG_ARGISNULL(5))
+		elog(ERROR, "pgfadvise_loader: databit argument shouldn't be NULL");
+	else
+		databit		= PG_GETARG_VARBIT_P(5);
+
 	/* initialize nulls array to build the tuple */
 	memset(nulls, 0, sizeof(nulls));
 
@@ -721,12 +742,16 @@ pgfincore_file(char *filename, pgfincoreStruct *pgfncr)
 	bits8	x = 0;
 	register int64 pageIndex;
 
+
 	/*
-	 * We work directly with the file
-	 * we don't use the postgresql file handler
+	 * We use the AllocateFile(2) provided by PostgreSQL.  We're going to
+	 * close it ourselves even if PostgreSQL close it anyway at transaction
+	 * end.
 	 */
-	struct stat	  st;
-	int			  fd;
+	FILE	*fp;
+	int	fd;
+	struct stat st;
+
 	void 		  *pa  = (char *) 0;
 	unsigned char *vec = (unsigned char *) 0;
 
@@ -743,15 +768,19 @@ pgfincore_file(char *filename, pgfincoreStruct *pgfncr)
 	pgfncr->rel_os_pages	= 0;
 
 	/*
-	* Open, fstat file
-	*/
-	fd = open(filename, O_RDONLY);
-	if (fd == -1)
-		return 1;
+	 * Fopen and fstat file
+	 * fd will be provided to posix_fadvise
+	 * if there is no file, just return 1, it is expected to leave the SRF
+	 */
+	fp = AllocateFile(filename, "rb");
+	if (fp == NULL)
+                return 1;
+
+	fd = fileno(fp);
 
 	if (fstat(fd, &st) == -1)
 	{
-		close(fd);
+		FreeFile(fp);
 		elog(ERROR, "Can not stat object file : %s",
 		     filename);
 		return 2;
@@ -770,7 +799,7 @@ pgfincore_file(char *filename, pgfincoreStruct *pgfncr)
 		pa = mmap(NULL, st.st_size, PROT_NONE, MAP_SHARED, fd, 0);
 		if (pa == MAP_FAILED)
 		{
-			close(fd);
+			FreeFile(fp);
 			elog(ERROR, "Can not mmap object file : %s, errno = %i,%s\nThis error can happen if there is not enought space in memory to do the projection. Please mail cedric@villemain.org with '[pgfincore] ENOMEM' as subject.",
 			     filename, errno, strerror(errno));
 			return 3;
@@ -781,7 +810,7 @@ pgfincore_file(char *filename, pgfincoreStruct *pgfncr)
 		if ((void *)0 == vec)
 		{
 			munmap(pa, st.st_size);
-			close(fd);
+			FreeFile(fp);
 			elog(ERROR, "Can not calloc object file : %s",
 			     filename);
 			return 4;
@@ -792,7 +821,7 @@ pgfincore_file(char *filename, pgfincoreStruct *pgfncr)
 		{
 			free(vec);
 			munmap(pa, st.st_size);
-			close(fd);
+			FreeFile(fp);
 			elog(ERROR, "mincore(%p, %lld, %p): %s\n",
 			     pa, (long long int)st.st_size, vec, strerror(errno));
 			return 5;
@@ -849,7 +878,7 @@ pgfincore_file(char *filename, pgfincoreStruct *pgfncr)
 	 */
 	free(vec);
 	munmap(pa, st.st_size);
-	close(fd);
+	FreeFile(fp);
 
 	/*
 	 * OS things : Pages free
